/*******************************************************************************
                    FIXFOXMF - Fix Foxpro Memo File
        (C) Copyright 1991, Quarterbyte Systems, Inc.  See Note 1 below.
--------------------------------------------------------------------------------
When a client computer crashed a few weeks ago, it left a memo file in an
unusable state because the FoxPro(R) file header gave a next-free-block address
larger than the file size would have allowed; Foxpro therefore claimed that the
memo file was not valid.  The situation came about because the file header had
been updated, but the new data records were still held in FoxPro's file cache.
This situation could arise on any standalone system using FoxPro, or any time
a database file opened for exclusive use on a network.

This program repairs the problem by correcting the memo file header.  After
this, the file will be usable, although it will most likely be missing some
data.  Also, the corresponding database file may contain references to memo file
blocks which were not written.  This is OK until you modify memo fields.  Then
the database file may contain duplicate and/or invalid memo block references.
This will lead to big trouble.

So, after using fixfoxmf on a memo file, immediately use the FoxPro "COPY"
command (with DELETED set OFF if you want to preserve deleted records) to copy
the database to a new file.  You can then delete the old file and rename the
new one.  Then rebuild any index files open at the time of the crash.

Usage:
	fixfoxmf [-v] [-t] [-n#] filename

where "filename" is the name of the memo file (e.g. VISIT.FPT).
Filename may have wild card characters (e.g "fixfoxmf *.fpt").

By default, the program will compute the correct address automatically and
will write the changes to the file.

The -v ("verbose") flag prints information to the standard output as the program
computes the correct address(es).

The -t ("test") flag inhibits the writing of the corrections to the file; the
program just computes the correct address.

The -n# ("number") flag sets the next-free-block address to # regardless of
whatever the program computes (e.g. -n1232).   Note: the -n flag may NOT be
used if the filename contains wildcards.

Notes:
--------------------------------------------------------------------------------
(1)  This program is copyrighted by Quarterbyte Systems, Inc. of Berkeley, CA.
     Permission is granted to the public to use, copy, and distribute this
     program with the proviso that the source and/or compiled object code and/or
     executable code derived from this program may be freely copied and given
     away but NOT sold, and that this copyright notice must be included with
     any copies.

     Furthermore, no guarantee is made of this program's suitability for any
     purpose.  Use is at your own risk.  Quarterbyte Systems, Inc. assumes no
     liability for any direct or indirect consequences of the use of this
     program.

(2)  FoxPro is a trademark of Fox Software, Inc.

(3)  To compile using Microsoft C, use the command line:
     cl     -Grs -Ozax fixfoxmf.c         to make fixfoxmf.exe (14.2K)
or   cl -AT -Grs -Ozax fixfoxmf.c         to make fixfoxmf.com (13.6K)

Implementation notes:
--------------------------------------------------------------------------------
The program determines the next free block address by examining the file header:

		      file length+block size-1
	next block = -------------------------
                             block size


I invite anyone who's interested to extend this program.  If a special
'fix the database file too' switch is given, have the program scan the
corresponding database file to make sure that it contains only valid memo file
references.  You need to scan the database file header to identify all memo
fields.  Then scan all the records.  If 'n' is a nonzero block number in any
memo field in the database, then you should be able to do this:

	compute: offset = block_length*n;

	verify:  readlong(fd, offset) = 0 or 1

	compute: memo_length = readlong(fd, offset+4);

	verify:  (offset+memo_length+8) <= filelength

If the block reference fails, write zero (10 blanks, ascii 32's) back to the
database file (unless the -t switch was specified).

*******************************************************************************/

char * copyright = "(C) Copyright 1991, Quarterbyte Systems, Inc. \
May be used and copied but not sold.";

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dos.h>
#include <io.h>

#define TRUE 1
#define FALSE 0

void bail         (char *);             // forward definitions
int  haswild      (char *);
char * mergepath  (char *, char *);
void process      (char *);
unsigned readword (FILE *, long);
long readlong     (FILE *, long);
int  writelong    (FILE *, long, long);
void wordrev      (char *);
void longrev      (char *);


int user_number = FALSE;		// TRUE if -n flag given
int verbose     = FALSE;		// TRUE if -v flag given
int test        = FALSE;		// TRUE if -t flag given
long new_nblk;				// new address to write out

char * usestr = "Usage: fixfoxmf [-v] [-t] [-n#] filename";

////////////////////////////////////////////////////////////////////////////////
// main - main routine
////////////////////////////////////////////////////////////////////////////////

main (argc, argv)
int argc;
char **argv;
{
    char * filename = NULL;		// pointer to database filename
    char * arg;				// pointer to argument string
    int didmatch = FALSE;		// true if wildcard matched >= 1 file
    struct find_t f;

    while (--argc > 0) {		// examine each argument
	if (**++argv == '-') {		// if it starts with '-' it's a switch
	    arg = strupr(*argv+1);	// convert string to upper case
	    while (*arg) {		// and examine each character
		switch (*arg++) {
		    case 'V':		// set verbose flag
			verbose = TRUE;
			break;

		    case 'N':		// set record count directly
			if (sscanf(arg, "%ld", &new_nblk) != 1)
			    bail(usestr); // error parsing the number

			user_number = TRUE;
			*arg = '\0';	// stop looking at this argument
			break;

		    case 'T':		// set test flag
			test = TRUE;
			break;

		    default:
			bail(usestr);
		}
	    }
	}
	else if (filename == NULL)	// first non - argument is the filename
	    filename = *argv;
	else
	    bail(usestr);		// there should not be two, however
    }

    if (filename == NULL)		// there should have been a name
	bail(usestr);

    if (haswild(filename)) {		// wildcards - search & process
	if (user_number)
	    bail("Can't specify -n flag with wildcards in filename");

	if (_dos_findfirst(filename, _A_NORMAL|_A_RDONLY, &f) == 0) {
	    do {			// process all matches
		process(mergepath(filename, f.name));
	    } while (_dos_findnext(&f) == 0);
	    didmatch = TRUE;
	}

	if (! didmatch) {		// no matches?
	    printf("%s: no matching filenames\n", filename);
	    exit(1);
	}
    }
    else
	process(filename);
}

////////////////////////////////////////////////////////////////////////////////
// haswild - return TRUE if filename as wildcard chars * or ?, FALSE if not
////////////////////////////////////////////////////////////////////////////////

int haswild (filename)
char * filename;
{
    return (strchr(filename, '?') != NULL) || (strchr(filename, '*') != NULL);
}

////////////////////////////////////////////////////////////////////////////////
// mergepath - append 'name' onto path part of 'path' and return the result
////////////////////////////////////////////////////////////////////////////////

char * mergepath (path, name)
char * path;
char * name;
{
    static char result[100];		// result string
    char * lastpchar;			// pointer to last path-type character
    char * p;				// pointer to a character

    lastpchar = strrchr(path, ':');	// look for a :, then a '\'
    if ((p = strrchr(path, '\\')) != NULL) {
	if (lastpchar == NULL)		// if just a '\'
	    lastpchar = p;		// take that
	else if (p > lastpchar)		// otherwise if both, take last one
	    lastpchar = p;
    }

    if (lastpchar == NULL)		// if no path characters
	result[0] = '\0';		// make result the empty string, else
    else				// copy up to & including the path char
	strncpy(result, path, lastpchar-path+1);

    strcat(result, name);		// append the name
    return result;
}

////////////////////////////////////////////////////////////////////////////////
// process - process one file ('filename')
////////////////////////////////////////////////////////////////////////////////

void process (filename)
char * filename;
{
    long file_length;			// size of file in bytes
    unsigned block_length;		// length of record in bytes
    long orig_nblk;			// original # of records in header
    long proper_nblk;			// number of records we compute
    FILE * fd;				// file descriptor

    if ((fd = fopen(filename, test ? "rb" : "r+b")) == NULL) {
	perror(filename);		// oops, can't
	exit(1);
    }

    printf("%-12s: %s", filename, verbose ? "\n" : "");

    file_length = filelength(fileno(fd)); // get file size in bytes
    if (verbose)
	printf("    file size:           %8ld bytes\n", file_length);

    if (file_length < 511L) {
	if (verbose)
	    printf("    file is too small\n");
	printf("not a FoxPro memo file\n");
	fclose(fd);
	return;
    }

    orig_nblk = readlong(fd, 0L);	// next-block value is at offset 0
    if (verbose)
	printf("    original next block: %8ld\n", orig_nblk);

    block_length = readword(fd, 6L);	// block size is at offset 6
    if (verbose)
	printf("    block size:          %8u\n", block_length);

    if (block_length < 32) {
	if (verbose)
	    printf("    block length is invalid\n");
	printf("not a FoxPro memo file\n");
	fclose(fd);
	return;
    }

    orig_nblk = readlong(fd, 0L);	// next-block value is at offset 0
    if (verbose)
	printf("    original next block: %8ld\n", orig_nblk);

	// compute record length from offset & length of the last field:
    proper_nblk = (file_length+block_length-1) / block_length;
    if (verbose)
	printf("    correct next block:  %8ld\n", proper_nblk);

    if (! user_number)			// set output number if not user-spec.
	new_nblk = proper_nblk;

    if (verbose)			// now show the results
	printf("    ");

    if (orig_nblk == new_nblk)		// if no change, say so
	printf("no change            %8ld\n", orig_nblk);
    else if (test)			// just say what we would have done
	printf("would change         %8ld -> %8ld\n", orig_nblk, new_nblk);
    else {
	if (! writelong(fd, 0L, new_nblk))
	    perror("error writing to file");
	else
	    printf("changed              %8ld -> %8ld\n", orig_nblk, new_nblk);
    }
    fclose(fd);
}

////////////////////////////////////////////////////////////////////////////////
// bail - print error message and exit
////////////////////////////////////////////////////////////////////////////////

void bail (str)
char * str;
{
    printf("%s\n", str);
    exit(1);
}

////////////////////////////////////////////////////////////////////////////////
// readword - read one word at offset 'offset'
////////////////////////////////////////////////////////////////////////////////

unsigned readword (fd, offset)
FILE *fd;
long offset;
{
    unsigned val;

    if (fseek(fd, offset, SEEK_SET) != 0)
	bail("seek error");

    if (fread(&val, 2, 1, fd) != 1)
	bail("read error");

    wordrev((char *) &val);
    return val;
}

////////////////////////////////////////////////////////////////////////////////
// readlong - read one long at offset 'offset'
////////////////////////////////////////////////////////////////////////////////

long readlong (fd, offset)
FILE *fd;
long offset;
{
    long val;

    if (fseek(fd, offset, SEEK_SET) != 0)
	bail("seek error");

    if (fread(&val, 4, 1, fd) != 1)
	bail("read error");

    longrev((char *) &val);
    return val;
}

////////////////////////////////////////////////////////////////////////////////
// writelong - write one long at offset 'offset'
////////////////////////////////////////////////////////////////////////////////

int  writelong(fd, offset, val)
FILE *fd;
long offset;
long val;
{
    if (fseek(fd, offset, SEEK_SET) != 0)
	bail("seek error");

    longrev((char *) &val);
    return fwrite(&val, 4, 1, fd) == 1;
}

////////////////////////////////////////////////////////////////////////////////
// longrev - byte-flip a long integer
////////////////////////////////////////////////////////////////////////////////

void longrev (v)
char *v;
{
   char tmp;

   tmp  = v[0];
   v[0] = v[3];
   v[3] = tmp;

   tmp  = v[1];
   v[1] = v[2];
   v[2] = tmp;
}

////////////////////////////////////////////////////////////////////////////////
// wordrev - byte-flip a word
////////////////////////////////////////////////////////////////////////////////

void wordrev (v)
char *v;
{
   char tmp;

   tmp  = v[0];
   v[0] = v[1];
   v[1] = tmp;
}
